(* ajouter open Pcf  l'extrieur de la bibliothque *)
open Op
open O
open Ast

type value =
  | Num_v of int
  | Rec_v of (Ast.label*value) list
  | Clo of string * Ast.t * env

and env = (string * value) list

let rec print_value chan v = match v with
  | Num_v i -> Printf.fprintf chan "%i" i
  | Rec_v l ->
      let rec iter o = function
        | [] -> ()
        | [lbl,v] -> Printf.fprintf o " %s=%a " lbl print_value v
        | (lbl,v)::l -> Printf.fprintf o " %s=%a ;%a" lbl print_value v iter l
      in
        Printf.fprintf chan "{%a}" iter l
  | Clo (_,_,_) -> Printf.fprintf chan "<fun>"

exception Error of string

let get_label lbl l =
  try
    List.assoc lbl l
  with
  | Not_found ->
      raise (Error (Printf.sprintf "Cannot access field %S" lbl))

let copy_with r l =
  (** On pourrait se contenter de rajouter les nouvelles associations en
   * tte de liste, mais je prends la peine d'viter les doublons,
   * mme si de faon peu efficace. Et puis, les doublons seraient affichs,
   * ce qui est moche. En fait, reprsenter les enregistrements par des Set
   * serait plus judicieux. *)
  List.fold_left
    (fun r' (lbl,v) ->
      if List.mem_assoc lbl r' then
        (lbl,v)::(List.remove_assoc lbl r')
      else
        (lbl,v)::r')
    r l
    
 
let rec interv env t = match t with
| Num i -> Num_v i
| Var x ->
    begin try
      List.assoc x env
    with Not_found ->
      raise (Error ("variable: "^x^" undefined")) end
| Op (op,t1,t2) ->
    let n1 = inter_int env t1 in
    let n2 = inter_int env t2 in
    Num_v
      (match op with
      | Add -> n1+n2
      | Sub -> let m = n1-n2 in if m < 0 then 0 else m
      | Mul -> n1*n2
      | Div -> n1/n2)
| Ifz (t1,t2,t3) ->
    let v1 = inter_int env t1 in
    interv env (if v1 = 0 then t2 else t3)
| Let (x,t1,t2) ->
    let v1 = interv env t1 in
    interv ((x,v1)::env) t2
| App (t1,t2) ->
    let x,t_clo,e_clo = inter_clo env t1 in
    let v2 = interv env t2 in
    interv ((x,v2)::e_clo) t_clo
| Fun (x,t) -> Clo (x,t,env)
| Fix (f,Fun (x,t)) ->
    let rec clo = Clo (x,t,(f,clo)::env) in
    clo
| Fix _ -> raise (Error "Fix allowed on Fun only")
(** Les enregistrements *)
| Rec l -> Rec_v (List.map (fun (l,e) -> l, interv env e) l)
| Access (r,lbl) ->
    let l = inter_rec env r in
    get_label lbl l
| CopyWith (r,l) ->
    let r = inter_rec env r in
    let l = List.map (fun (lbl,t) -> lbl, interv env t) l in
    Rec_v (copy_with r l)
(** Objects *)
(* Les deux cas suivants sont traits selon la smantique *)
| Obj (xo,l) ->
    let x = match xo with  None -> "this" | Some x -> x in
    Rec_v (List.map (fun (lbl,t) -> lbl,Clo (x,t,env)) l)
| Invoke (t,lbl) ->
    let o = interv env t in
    let l = as_rec o in
    let (x,tc,ec) = as_clo (get_label lbl l) in
    interv ((x,o)::ec) tc
(** Ici on interprte ObjWith comme du  sucre syntaxique  *)
| ObjWith (xo,r,l) ->
    let x = match xo with  None -> "this" | Some x -> x in
    interv env
      (Let ("super",r,
            CopyWith
              (Var "super",
               List.map (fun (lbl,t) -> lbl,Fun (x,t)) l)))

and inter_int env t = match interv env t with
| Num_v i -> i
| Clo _ -> raise (Error "Num_v expected, got Clo")
| Rec_v _ -> raise (Error "Num_v expected, got Rec_v")

and inter_clo env t = as_clo (interv env t)

and as_clo v = match v with
| Clo (x,t,e) -> (x,t,e)
| Num_v _ -> raise (Error "Clo expected, got Num_v")
| Rec_v _ -> raise (Error "Clo expected, got Rec_v")

and inter_rec env t = as_rec (interv env t)

and as_rec v = match v with
| Rec_v l -> l
| Clo _ -> raise (Error "Rec expected, got Clo")
| Num_v _ -> raise (Error "Rec expected, got Num_v")

(* Exemple

let v = obj self
    {orchestre =  150;
     balcon = 30;

     reserveo =
        obj { self with orchestre = self#orchestre - 1 } ;     
     annuleo = 
        obj { self with orchestre  = self#orchestre + 1 } ;

     reserveb =
        obj { self with balcon = self#balcon - 1 } ;     
     annuleb = 
        obj { self with balcon  = self#balcon + 1 } ;
}
;;

v#reserveo#reserveo#annuleo#reserveo#reserveb#orchestre
;;

let v =
  obj self
    { { orchestre =  150;  balcon = 30; }
     with
     reserveo =
        { self with orchestre = self.orchestre - 1 } ;     
     annuleo = 
        { self with orchestre  = self.orchestre + 1 } ;

     reserveb =
        { self with balcon = self.balcon - 1 } ;     
     annuleb = 
         { self with balcon  = self.balcon + 1 } ;
} ;;

v#reserveo#reserveo#annuleo#reserveo#reserveb.orchestre
;;

*)
